const PLAYER_MAX_SPEED = 96.0;
const PLAYER_TURN_SPEED = Math.PI;
const CAMERA_WIDTH = 256.0;
const CAMERA_LENGTH = 64.0;
const CAMERA_FAR = 6.0; //distance of 6 * 64 in world coordinates
var gForwardKey = 87; //w
var gBackwardKey = 83; //s
var gStrafeLeftKey = 65; //a
var gStrafeRightKey = 68; //d
var gTurnLeftKey = 37; //left
var gTurnRightKey = 39; //right
var gFireKey = 32; //space
var gBendKey = 81; //q
const PLAYER_DEF = 0;
const PLAYER_DEAD = 1;
const MAX_CRYSTALS = 25;
function Player(x, y, a, r)
{
    Entity.call(this, x, y, a, r);
    this.z = 0.0;
    this.fwd = 0.0;
    this.stf = 0.0;
    this.cosa = Math.cos(a);
    this.sina = Math.sin(a);
    this.icosa = Math.cos(-a);
    this.isina = Math.sin(-a);
    this.noclip = false;
    this.walk = 0;
    this.fireTimer = 0;
    this.health = 100;
    this.state = PLAYER_DEF;
    this.crystals = 0;
    this.bending = false;
    this.flameSound = new Audio("flamethrower.wav");
}
Player.prototype = 
{
    canHurt: true,
    playerType: true,
    update: function(deltaTime)
    {  
        if (this.state == PLAYER_DEF)
        {
            //Cheaty cheats
            if (gKeyHits[106]) this.noclip = !this.noclip; //Numpad *
            if (gKeyHits[96]) //Numpad 0
            {
                var fr = new Friend(this.x + this.cosa * 20.0, this.y + this.sina * 20.0);
                fr.frameY = 176;
                gGameState.addEnt(fr);
            }
            if (gKeyHits[109]) this.hurt(50); //Numpad -
            if (gKeyHits[107]) this.health = 100; //Numpad +
            if (gKeyHits[105]) this.crystals+=5; //Numpad 9
            if (gKeyStatus[98]) //Numpad 2
            {
                for (var i = 0; i < gGameState.mapEntities.length; i++)
                {
                    if (gGameState.mapEntities[i].treeType)
                    {
                        gGameState.mapEntities[i].growthStage = 0;
                        gGameState.mapEntities[i].solid = false;
                        gGameState.mapEntities[i].updateFrame();
                    }
                }
            }

            if (this.crystals >= MAX_CRYSTALS && gKeyHits[gBendKey] && !this.bending)
            {
                this.bending = true;
            }
            if (this.bending)
            {
                this.crystals -= deltaTime;
                if (this.crystals <= 0.0)
                {
                    this.crystals = 0;
                    this.bending = false;
                }
                for (var i = 0; i < gGameState.mapEntities.length; i++)
                {
                    if (gGameState.mapEntities[i].treeType)
                    {
                        var tree = gGameState.mapEntities[i];
                        var sqDist = distance2DSqr(tree.x, tree.y, this.x, this.y);
                        if (sqDist < 9216.0) 
                        {
                            tree.newZ = -32.0;
                        }
                    }
                }
            }

            if (gKeyStatus[gForwardKey])
            {
                this.fwd = PLAYER_MAX_SPEED;
                this.walk += deltaTime;
            } 
            else if (gKeyStatus[gBackwardKey])
            {
                this.fwd = -PLAYER_MAX_SPEED;
                this.walk -= deltaTime;
            }
            else
            {
                this.fwd = 0.0;
            }

            if (gKeyStatus[gStrafeLeftKey])
            {
                this.stf = -PLAYER_MAX_SPEED;
                this.walk += deltaTime * (Math.sign(this.fwd) || 1.0);
            }
            else if (gKeyStatus[gStrafeRightKey])
            {
                this.stf = PLAYER_MAX_SPEED;
                this.walk += deltaTime * (Math.sign(this.fwd) || 1.0);
            }
            else
            {
                this.stf = 0.0;
            }

            if (gKeyStatus[gTurnLeftKey])
            {
                this.a -= PLAYER_TURN_SPEED * deltaTime;
            }
            else if (gKeyStatus[gTurnRightKey])
            {
                this.a += PLAYER_TURN_SPEED * deltaTime;
            }

            if (gKeyStatus[110]) //numpad period
            {
                this.fwd = PLAYER_MAX_SPEED * 4.0;
                this.walk += deltaTime * 4.0;
            }

            if (gKeyStatus[gFireKey])
            {
                this.fireTimer += deltaTime;
                if (this.fireTimer > 0.01)
                {
                    this.fireTimer = 0.0;
                    gGameState.addEnt(new Fireball(this.x + this.cosa * this.r, this.y + this.sina * this.r, 4.0, this.a, this));
                    if (this.flameSound.currentTime > 0.1 || this.flameSound.paused)
                    {
                        this.flameSound.play();
                        this.flameSound.currentTime = 0;
                    }
                }
            }
            else
            {
                this.fireTimer = 0;
            }

            this.cosa = Math.cos(this.a);
            this.sina = Math.sin(this.a);
            this.icosa = Math.cos(-this.a);
            this.isina = Math.sin(-this.a);
            
            var dx = this.cosa * this.fwd * deltaTime - this.sina * this.stf * deltaTime;
            var dy = this.sina * this.fwd * deltaTime + this.cosa * this.stf * deltaTime;
            if (this.noclip)
            {
                this.x += dx;
                this.y += dy;
            }
            else
            {
                this.tryMove(dx, dy);
            }
        }
        else if (this.state == PLAYER_DEAD)
        {
            this.z += deltaTime * 8.0;
            if (this.z > 14.0) this.z = 14.0;
        }
    },
    hurt: function(amt)
    {
        if (this.state != PLAYER_DEAD)
        {
            this.health -= amt;
            gGameState.flashScreen("255,0,0", 0.4);
            if (this.health <= 0)
            {
                this.health = 0;
                this.state = PLAYER_DEAD;
                gGameState.fadeScreen("255, 0, 0", 5.0);
                gGameState.loseTimer = 5.0;
                PlaySound("dilan_death_" + Math.floor(Math.random() * 2.0) + ".wav");
            }
            else
            {
                PlaySound("dilan_hurt_" + Math.floor(Math.random() * 4.0) + ".wav");
            }
        }
    },
    render2D: function(x, y)
    {
        gContext.fillStyle = "#FF0000";
        gContext.beginPath();
        gContext.arc(x, y, this.r, 0.0, TWOPI, false);
        gContext.fill();
        gContext.strokeStyle = "#AA0000";
        gContext.beginPath();
        gContext.moveTo(x, y);
        gContext.lineTo(x + this.cosa * this.r * 2.0,
            y + this.sina * this.r * 2.0);
        gContext.stroke();
    },
    render3D: function(player)
    {
        //Nope
    },
    toCameraSpace: function(wx, wy, wz, out)
    {
        /*
        [-sina cosa 0]   [x]   [x * -sina + y * cosa]
        [cosa sina  0] * [y] = [x * cosa + y * sina]
        [0    0     1]   [1]   [0 + 0 + 1]
        */
       wx -= this.x;
       wy -= this.y;
       wz -= this.z;
       out = out || [];
       out[0] = wx * this.isina + wy * this.icosa;
       out[1] = wx * -this.icosa + wy * this.isina;
       out[2] = wz; //No pitch rotation, so no need to transform z
       return out;
    },
    toClipSpace: function(cx, cy, cz, out)
    {
        out = out || [];
        out[0] = cx / CAMERA_WIDTH;
        out[1] = -cy / CAMERA_LENGTH;
        out[2] = cz / (CAMERA_WIDTH / gAspectRatio);
        out[3] = -cy; //We need an unscaled depth value for more accurate sorting.
        return out;
    }
}
Object.setPrototypeOf(Player.prototype, Entity.prototype);